/**
 * Implementaci del protocol HTTP
 *
 * Autor: Miquel Fontanals Rofes
 */

#include "main.h"

#define HTTP11 (strcmp (clients[cid].dades->versio, "HTTP/1.1") == 0)
#define TEHOST (strlen (clients[cid].dades->Host) > 0)
#define CGI ((strcmp (clients[cid].dades->metode, "POST") == 0) || (strlen (clients[cid].dades->get_p) > 0))
#define POSTOK ((strcmp (clients[cid].dades->metode, "POST") == 0) && (clients[cid].dades->ContentLength > 0))


/* Buffer pels logs */
char log_temp[2048];

/**
 * Funci auxiliar strcat
 *
 * Parmetres:
 * 		- *sbuf: el buffer de sortida
 * 		- *s: el que li volem afegir
 */
void xic_strcat (buf_out *sbuf, char *s)
{
	int i, mida_s;
	mida_s = strlen (s);
	for (i = 0; i < mida_s; i++) {
		(*sbuf).sbuf[(*sbuf).mida] = s[i];
		(*sbuf).mida++;
	}
}

/**
 * Funci auxiliar strncat.
 *
 * Parmetres:
 * 		- *sbuf: el buffer de sortida
 * 		- *s: el que li volem afegir
 * 		- mida_s: la mida de *s
 *
 * NOTA: Donat que el XicHttpd sempre que pot omple el buffer de sortida
 * per enviar la resposta al client ens sorgeix un problema quan ens
 * demanen fitxers binaris, o quan es transfereix en gzip, en aquests casos
 * les funcions strcat i strncat de la llibreria del C s'aturen en trobar
 * el primer '\0'. Les funcions strlen tampoc ens donen els resultats que
 * esperem, per tant es defineix l'estructura buf_out (http.h) per tal de mantenir
 * la mida coerent.
 */
void xic_strncat (buf_out *sbuf, char *s, int mida_s)
{
	int i = 0;
	for (i = 0; i < mida_s; i++) {
		(*sbuf).sbuf[(*sbuf).mida] = s[i];
		(*sbuf).mida++;
	}
}

/**
 * Funci que retorna la data actual seguint el format definit
 * a l'RFC1123
 *
 * Retorn:
 * 		- char*: La data actual
 */
char *getData ()
{
	time_t ara;
	static char timebuf[100];
	ara = time ((time_t *)0);
	strftime (timebuf, sizeof (timebuf), RFC1123, gmtime(&ara));
	return (char *) timebuf;
}

/**
 * Funci que decodifica un missatge de petici
 * 
 * Parmetre:
 * 		- cid: Identificador del client, que correspn a l'ndex
 * 		  de la posici que ocupa de la taula de clients
 * 		  
 * Retorn:	-1 - No hi ha res per llegir
 * 		0 - sense errors
 *		CE400 - Bad Request general
 *		CE501 - No implementat
 *		CE414 - URI massa llarga
 *		CE505 - Versio HTTP no soportada
 *		CE411 - Un POST sense Content-Length
 */
int deco_request (int cid)
{
	char buf[2048];
	char *ptr[2048];	/* strtok_r */
	char *aux, *val, *get_params;
	int n, ret = 0;

	/* Lnia de comanda */
	if ((n = llegir_linia (clients[cid].socket, &buf[0], 2048)) > 0) {
		strncpy (log_temp, buf, n - 1);
		log_temp[n - 1] = '\0';

		if ((aux = (char *) strtok_r (buf, " \t", ptr)) == NULL) return CE400;
		/* Primer el mtode */
		if ((strcmp (aux, "GET") != 0) && (strcmp (aux, "HEAD") != 0) && (strcmp (aux, "POST") != 0))
			return CE501;
		sprintf (clients[cid].dades->metode, "%s", aux);
		/* L'URI */
		if ((aux = (char *) strtok_r (NULL, " \t", ptr)) == NULL) return CE400;
		if (strlen (aux) > 1500)
			return CE414;
		/* Query */
		if ((get_params = (char *) strchr (aux, '?')) != NULL) {
			*get_params = '\0';
			get_params++;
			if (strlen (get_params) > 0) {
				sprintf (clients[cid].dades->get_p, "%s", get_params);
#ifdef DEBUG
				fprintf (f_err, "Get params: %s\n", clients[cid].dades->get_p);
				fprintf (f_err, "URI: %s\n", aux);
#endif
			}
		}
		sprintf (clients[cid].dades->URI, "%s", aux);
		if (strcmp (clients[cid].dades->URI, "/") == 0)
			strcat (clients[cid].dades->URI, "index.html");

		/* La versi HTTP */
		if ((aux = (char *) strtok_r (NULL, " \t\r\n", ptr)) == NULL) return CE400;
		if ((strcmp (aux, "HTTP/1.0") != 0) && (strcmp (aux, "HTTP/1.1") != 0))
			return CE505;
		sprintf (clients[cid].dades->versio, "%s", aux);
	} else {
		char err[64];
		sprintf (err, "%i", cid);
		fprintf (f_err, FORMAT_ERROR, getData (), XREQUEST,
				"HTTP->DECO_REQUEST", err);
		return -1;
	}
	/* Les capaleres, descartem les que no ens interessen, retornem CE400 per les capaleres mal
	 * construides */
	while ((n = llegir_linia (clients[cid].socket, &buf[0], 2048)) > 0) {
		if ((aux = (char *) strchr (buf, ':')) == NULL) ret = CE400;
		else if (((aux = (char *) strtok_r (buf, ": \t", ptr)) == NULL) ||
			((val = (char *) strtok_r (NULL, " \r\n", ptr)) == NULL)) ret = CE400;
		if (ret == 0) {
			/* Host */
			if (strcmp (aux, "Host") == 0) {
				if (strlen (val) < 64)
					sprintf (clients[cid].dades->Host, "%s", val);
				else ret = CE400;
			/* Accept-Encoding */
			} else if ((strcmp (aux, "Accept-Encoding") == 0) && (((char *) strstr (val, "gzip")) != NULL))
				sprintf (clients[cid].dades->AcceptEncoding, "gzip");
			/* Content-Length per als POST */
			else if ((strcmp (aux, "Content-Length") == 0) && (strlen (val) <= 5))
				clients[cid].dades->ContentLength = atoi (val);
			/* Connection */
			else if ((strcmp (aux, "Connection") == 0) && (strlen (val) < 64))
				sprintf (clients[cid].dades->Connection, "%s", val);
		}
	}
	/* Si s HTTP/1.1 comprovem que tingui la capalera Host */
	/* i si s un POST que tingui Content-Length */
	if ((HTTP11 && !TEHOST)) ret = CE400;
	if (CGI) {
		if ((strcmp (clients[cid].dades->metode, "POST") == 0) &&
				(clients[cid].dades->ContentLength == 0))
			ret = CE411;
	}

	return ret;
}

/**
 * Funci que llegeix una lnia del socket passat per parmetre i en deixa el contingut
 * a buf, sempre hi quan no rebassi la mida del mateix.
 * 
 * Parmetres:	
 * 		sock - El socket d'on llegirem
 * 		*buf - El buffer on es deixar la lnia llegida
 * 		mida - La mida mxima que pot emmagatzemar buf
 *
 * Retorn:	0  - No hem llegit res
 *		-1 - Timeout exhaurit
 *		>0 - El nombre de bytes llegits
 *
 * Nota: La lectura es fa carcter a carcter fins que trobem un \r
 * 	 Un 0 de retorn tamb indica una lnia en blanc, s a dir, el final de les capaleres
 */
int llegir_linia (int sock, char *buf, int mida)
{
	int i, n;
	char c = '\0';
	memset (buf, '\0', mida);
	for (i = 0; (i < (mida - 1)) && (c != '\n'); i++) {
		n = recv (sock, &c, 1, MSG_NOSIGNAL);
		if (n > 0) {
			if (c == '\r') {	// mirem si el segueix un \n
				n = recv (sock, &c, 1, MSG_PEEK);
				if ((n > 0) && (c == '\n'))
					recv (sock, &c, 1, MSG_NOSIGNAL);
				else c = '\n';	// ho forcem
			}
			buf[i] = c;
		} else c = '\n';
	}
	
	buf[i] = '\0';
	// debug
#ifdef DEBUG
	fprintf (f_err, "%s", buf);
#endif
	if (i == 1) return 0;	// Linea en blanc, fi de les capaleres

	return (i);			
}

/**
 * Funci que inicialitza les variables de l'estrucutra c_request d'aquest
 * client.
 *
 * Parmetres:	cid - L'identificador d'aquest client
 */
void reset_request (int cid)
{
	memset (clients[cid].dades->metode, '\0', 16);
	sprintf (clients[cid].dades->versio, "HTTP/1.1");
	memset (clients[cid].dades->URI, '\0', 1024);
	memset (clients[cid].dades->Host, '\0', 64);
	memset (clients[cid].dades->AcceptEncoding, '\0', 16);
	memset (clients[cid].dades->get_p, '\0', 1024);
	clients[cid].dades->ContentLength = 0;
	memset (clients[cid].dades->Connection, '\0', 64);
}

/**
 * Funci que afegeix la capalera Content-Type de la resposta en funci de l'extensi.
 * - Si extensi apunta a NULL s'afegeix el tipus MIME text/plain
 * - Si s una extensi desconeguda per aquesta funci, s'afegeix el tipus MIME application/octet-stream
 *
 * Parmetres:	*resposta - Punter a l'estructura que cont la resposta
 * 		*extensi - Punter a la extensi d'algun fitxer
 *
 */
void set_mime (c_response *resposta, char *extensio)
{
	mime *actual;
	if (extensio == NULL) {
		sprintf ((*resposta).ContentType, "Content-Type: text/plain\r\n");
		return;
	}
	for (actual = mime_types; actual->next != NULL; actual = actual->next) {
		if (strcmp (extensio, actual->ext) == 0) {
			sprintf ((*resposta).ContentType, "Content-Type: %s\r\n", actual->tipus_mime);
			return;
		}
	}
	// Si som aqu no hem trobat el tipus mime
	sprintf ((*resposta).ContentType, "Content-Type: application/octet-stream\r\n");
}

/**
 * Funci que envia les dades al client. El punter *sbuf contindr una o varies
 * respostes, i el punter *dades contindr les dades a enviar, que poden ser
 * un fitxer, o qualsevol altre cosa. Si el client est fent les peticions amb
 * pipelining, s'afegiran les dades al bufer de sortida *sbuf, quan *sbuf sigui ple
 * o b si el client no soporta pipelinig, s'enviar el contingut del buffer de
 * sortida directament pel socket.
 *
 * Parmetres:	sock		- El socket del client
 * 		*dades		- Punter a les dades que es volen enviar si s NULL
 * 				  es buida el buffer de sortida
 * 		mida_dades	- La mida de les dades
 * 		*sbuf		- Punter al buffer de sortida
 * 		mida_sbuf	- La capacitat mxima del buffer de sortida
 *
 * Retorn:	 0 - Si tot s correcte
 * 		-1 - Si es produeixen errors
 *
 * NOTA: Amb el parmetre MSG_NOSIGNAL de les crides a send ens estalviem
 * de rebre els SIGPIPE quan ens tanquen l'altre extrem del socket, aix
 * la funci retorna -1 amb el que finalitzem aquest thread.
 */
int envia (int sock, char *dades, int mida_dades, buf_out *sbuf, int mida_sbuf)
{
	int lliure, pft;
	
	/* s necessari restar-li tamb tot el que les dems funcions
	 * puguin escriure posteriorment al buffer de sortida */
	lliure = mida_sbuf - (*sbuf).mida - MIDA_MAX_RESPONSE;
	
#ifdef DEBUG
	write (2, (*sbuf).sbuf, (*sbuf).mida); // cal un write per treure els binaris
#endif

	if ((pft = comprova_pipelining (sock)) == -1) return -1;

	/* Primer, si ens hi cap al buffer de sortida, ho ajuntem
	 * tot abans d'enviar */
	if ((mida_dades < lliure) && (dades != NULL)) {
		xic_strncat (sbuf, dades, mida_dades);
		/* Pipelinning, ho enviarem quan sigui ple */
		if (pft == 1) return 0;
		/* Ho enviem inmediatament */
		else {
			if (send (sock, (*sbuf).sbuf, (*sbuf).mida, MSG_NOSIGNAL) < 0)
				return -1;
			(*sbuf).mida = 0;
			return 0;
		}
	/* Enviem el buffer de sortida i les dades seguidament */
	} else {
		if (send (sock, (*sbuf).sbuf, (*sbuf).mida, MSG_NOSIGNAL) < 0) return -1;
		(*sbuf).mida = 0;
		if (dades == NULL) return 0;
		if (send (sock, dades, mida_dades, MSG_NOSIGNAL) < 0) return -1;
		return 0;
	}
	return -1;	// No hi hauriem d'arribar
}

/**
 * Funci que prepara el missatge de resposta a un error.
 *
 * Parmetres:	ce	- El codi d'error
 * 		cid	- L'identificador d'aquest client
 * 		*sbuf	- Punter al buffer de sortida
 *
 * Retorna el que retorni la funci envia
 */
int prepara_error (int ce, int cid, buf_out *sbuf)
{
	c_response resposta;
	char html[1024];
	
	/* Preparem el missatge */
	sprintf (html, "<html><body><center><h1>");
	sprintf (resposta.Server, "Server: %s\r\n", NOM_SERVIDOR);
	sprintf (resposta.Date, "Date: %s\r\n", getData ());

	sprintf (resposta.estat, "%s ", clients[cid].dades->versio);

	/* Seleccionem l'error corresponent */
	switch (ce) {
		case CE400:
			strcat (resposta.estat, HTTP400);
			strcat (html, HTTP400);
			break;
		case CE501:
			strcat (resposta.estat, HTTP501);
			strcat (html, HTTP501);
			break;
		case CE414:
			strcat (resposta.estat, HTTP414);
			strcat (html, HTTP414);
			break;
		case CE505:
			strcat (resposta.estat, HTTP505);
			strcat (html, HTTP505);
			break;
		case CE404:
			strcat (resposta.estat, HTTP404);
			strcat (html, HTTP404);
			break;
		case CE403:
			strcat (resposta.estat, HTTP403);
			strcat (html, HTTP403);
			break;
		case CE411:
			strcat (resposta.estat, HTTP411);
			strcat (html, HTTP411);
			break;
		default:
			strcat (resposta.estat, HTTP500);
			strcat (html, HTTP500);
	}
	fprintf (f_log, LOG_FORMAT, inet_ntoa (clients[cid].adreca_client.sin_addr),
			getData (), log_temp, resposta.estat);
	
	/* Preparem el que ens falta */
	strcat (resposta.estat, "\r\n");
	strcat (html, "</h1></center></body></html>");
	sprintf (resposta.ContentLength, "Content-Length: %i\r\n", strlen (html));
	sprintf (resposta.ContentType, "Content-Type: text/html\r\n");
	
	/* Enviem el missatge */
	xic_strcat (sbuf, resposta.estat);
	xic_strcat (sbuf, resposta.Date);
	xic_strcat (sbuf, resposta.Server);
	xic_strcat (sbuf, resposta.ContentLength);
	xic_strcat (sbuf, resposta.ContentType);
	if ((!HTTP11) || (strcasecmp (clients[cid].dades->Connection, "close")) == 0)
		xic_strcat (sbuf, "Connection: close\r\n");

	/* Si Content-Length s major que 0 descartem l'entity-body
	 * per no confondre'l amb una connexi amb pipelining */
	if (clients[cid].dades->ContentLength > 0) { 
		char c;
		int i = 1;
		while ((recv (clients[cid].socket, &c, 1, MSG_NOSIGNAL) > 0) &&
				(i++ < clients[cid].dades->ContentLength)); 
		xic_strcat (sbuf, "\r\n");
		envia (clients[cid].socket, html, strlen (html), sbuf, MAX_SBUF);
		return envia (clients[cid].socket, NULL, 0, sbuf, MAX_SBUF); // forcem l'enviament
	}
	/* Enviem segons la petici */
	if (strcmp (clients[cid].dades->metode, "HEAD") == 0) {
		return envia(clients[cid].socket, "\r\n", strlen ("\r\n"), sbuf, MAX_SBUF);
	}
	xic_strcat (sbuf, "\r\n");
	return envia (clients[cid].socket, html, strlen (html), sbuf, MAX_SBUF);
	
}

/**
 * Funci que s'executar quan hi hagi una petici CGI que es
 * pugui atendre (veure cerca_recurs)
 *
 * Parmetres:
 * 		- cid: L'identificador d'aquest client
 * 		- sbuf: El buffer de sortida
 * 		- f_cgi: Nom de l'executable demanat
 * Retorn:
 * 		0 	: Tot correcte
 * 		-1	: Algun error inesperat
 * 		CE500	: Error en l'execuci del CGI (Internal Server Error)
 */
int xic_cgi (int cid, buf_out *sbuf, char *f_cgi)
{
	char cgi_buf[1024];
	int cgi_out[2];
	int cgi_in[2];
	pid_t pid;
	int estat;
	char c;

#ifdef DEBUG
	fprintf (f_err, "\nCGI: %s\n", f_cgi);
#endif
	
	sprintf (cgi_buf, "%s %s", clients[cid].dades->versio, HTTP200);

	/* Un cop hem preparat la sortida
	 * Reemplacem la versi HTTP per tancar la connexi (...) */
	sprintf (clients[cid].dades->versio, "HTTP/1.0");

	/* Creem dues pipes per la comunicaci bidireccional */
	if (pipe (cgi_out) < 0) return CE500;
	if (pipe (cgi_in) < 0) return CE500;
	
	/* Creem el nou procs */
	if ((pid = fork ()) < 0) return CE500;
	if (pid == 0) {		/* Fill - El CGI */
		char entorn_metode[255];
		char entorn_get_p[255];
		char entorn_mida[255];

		/* Dupliquem l'entrada i sortida que ens interessa
		 *  i tanquem la resta */
		dup2 (cgi_out[1], 1);
		dup2 (cgi_in[0], 0);
		close (cgi_out[0]);
		close (cgi_in[1]);

		/* Les variables d'entorn mnimes que necessiten els CGI */
		sprintf (entorn_metode, "REQUEST_METHOD=%s", clients[cid].dades->metode);
		sprintf (entorn_get_p, "QUERY_STRING=%s", clients[cid].dades->get_p);
		sprintf (entorn_mida, "CONTENT_LENGTH=%i", clients[cid].dades->ContentLength);
		putenv (entorn_metode);
		putenv (entorn_get_p);
		putenv (entorn_mida);

		/* Executem el CGI */
		execl (f_cgi, f_cgi, NULL);
		
		/* Si sm aqu alguna cosa no ha anat b */
		write (2, "\nEXECL\n", strlen ("\nEXECL\n"));
		exit (0);
	} else {		/* Pare */
		int i;

		/* Tanquem els canals que sobren */
		close (cgi_out[1]);
		close (cgi_in[0]);

		/* Si era un POST, passem l'entity-body al CGI */
		if (POSTOK) {
			for (i = 0; i < clients[cid].dades->ContentLength; i++) {
				if (recv (clients[cid].socket, &c, 1, MSG_NOSIGNAL) > 0)
					write (cgi_in[1], &c, 1);
			}
		}
		/* Preparem la resposta */
		xic_strcat (sbuf, cgi_buf);
		xic_strcat (sbuf, "\r\n");
		xic_strcat (sbuf, "Connection: close\r\n"); // Perqu l'altre extrem tamb tanqui
		i = 0; // Per detectar els errors
		/* Utilitzem el buffer de sortida sbuf per enviar
		 * la sortida del cgi */
		for (;;) {
			if ((*sbuf).mida >= MAX_SBUF - 1) {
				if (send (clients[cid].socket, (*sbuf).sbuf, (*sbuf).mida, MSG_NOSIGNAL) < 0) {
					i = -1;
					break;
				}
				(*sbuf).mida = 0;
			}
			if (read (cgi_out[0], &c, 1) > 0) {
				(*sbuf).sbuf[(*sbuf).mida] = c;
				(*sbuf).mida++;
			} else {
				if (send (clients[cid].socket, (*sbuf).sbuf, (*sbuf).mida, MSG_NOSIGNAL) < 0) {
					i = -1;
					break;
				}
#ifdef DEBUG
				fprintf (f_err, "%s", (*sbuf).sbuf);
#endif
				break;
			}
		}
	
		/* Tanquem el que queda */
		close (cgi_out[0]);
		close (cgi_in[1]);

		if (i == -1) return -1; // Perqu probablement ha fallat a l'enviar

		/* Esperem a que el CGI acabi */
		waitpid (pid, &estat, 0);
		if (WIFEXITED (estat)) {
			fprintf (f_log, LOG_FORMAT, inet_ntoa (clients[cid].adreca_client.sin_addr),
					getData (), log_temp, cgi_buf);
			return 0;
		} else return -1;			// Algun error estrany
	}
}

/**
 * Funci que s'encarrega d'enviar el recurs demanat comprimit amb gzip
 *
 * Parmetres:
 * 		- cid:		L'id d'aquesta connexi
 * 		- file_in:	El recurs que es vol comprimir
 * 		- mida_in:	La mida de file_in
 * 		- resposta:	El c_response que estem construint
 * 		- sbuf:		El buffer de sortida
 * Retorn:
 * 		-1:	Algun error
 * 		 0:	Tot correcte
 *
 * TODO: De moment es crea un fitxer temporal per a la compressi, per s'envia
 * directament de la memria enlloc de llegir-lo del disc.
 * - La grcia seria crear el fitxer comprimit completament en memria;
 * - o b saber la mida de sortida per omplir la capalera i desprs fer la 
 *   compressi directament cap al socket;
 * - o b reciclar els temporals ja creats implementant una cau.
 */
int xic_gzip (int cid, char *file_in, int mida_in, c_response *resposta, buf_out *sbuf)
{
	/* Tro d'estructura extreta de la llibreria zlib
	 * ja que necessitem saber la mida de sortida abans
	 * d'enviar el fitxer per omplir la capalera
	 * per est declarat com a "local" a gzio.c.
	 * Aix ens permet accedir al membre stream de l'estructura
	 */
	typedef struct xic_gz_stream {
		z_stream stream;
	} xic_gz_stream;
	
	int fdout, escrits, ret = 0, mida_ant;		// fd desti, ...
	FILE *fdin, *tmp;				// fitxer origen, temporal
	char buf_in[mida_in];				// buffer de lectura
	gzFile gzfile;					// fitxer gz
	caddr_t buf_mem;				// per mapejar a memoria
	
	/* Obrim l'original */
	if ((fdin = fopen (file_in, "r")) == NULL)
		return -1;
	/* Preparem el buffer per fer la compressi d'un sol cop */
	if (fread (buf_in, 1, mida_in, fdin) != mida_in)
		return -1;
	fclose (fdin);
	/* Creem el fitxer temporal */
	tmp = tmpfile ();
	fdout = fileno (tmp);
	/* Comencem la compressi */
	if (!(gzfile = gzdopen (fdout, "wb9")))
		return -1;
	if (!(gzwrite (gzfile, buf_in, mida_in)))
		return -1;
	if (gzflush (gzfile, Z_SYNC_FLUSH) != Z_OK)
		return -1;
	/* Escrivim la mida */
	escrits = ((xic_gz_stream *) gzfile)->stream.total_out + 20; // 20 = Capaleres + CRC
	/* Projectem el fitxer comprimit en memria ara que sabem quan ocupar */
	if ((buf_mem =  mmap ((void *) 0, escrits, PROT_WRITE | PROT_READ, MAP_SHARED, fdout, 0)) == (caddr_t)-1)
		return -1;
	/* Ja podem tancar el gzip */
	if (gzclose (gzfile) != Z_OK)
		return -1;

	/* Les capaleres que falten */
	sprintf ((*resposta).ContentLength, "Content-Length: %i\r\n", escrits);
	sprintf ((*resposta).ContentEncoding, "Content-Encoding: gzip\r\n");

	/* Guardem la mida anterior del buffer per si falla */
	mida_ant = (*sbuf).mida;
	/* Preparem la resposta */
	xic_strcat (sbuf, (*resposta).estat);
	xic_strcat (sbuf, (*resposta).Date);
	xic_strcat (sbuf, (*resposta).Server);
	xic_strcat (sbuf, (*resposta).LastModified);
	xic_strcat (sbuf, (*resposta).ContentType);
	xic_strcat (sbuf, (*resposta).ContentLength);
	xic_strcat (sbuf, (*resposta).ContentEncoding);
	if ((!HTTP11) || (strcasecmp (clients[cid].dades->Connection, "close") == 0))
		xic_strcat (sbuf, "Connection: close\r\n");

	/* Si s un HEAD noms enviem les capaleres
	 * Si s un GET li enviem les capaleres seguit d'un CRLF i del fitxer */
	if (strcmp (clients[cid].dades->metode, "HEAD") == 0) {
		ret = envia (clients[cid].socket, "\r\n", strlen ("\r\n"), sbuf, MAX_SBUF);
	} else if (strcmp (clients[cid].dades->metode, "GET") == 0) {
		xic_strcat (sbuf, "\r\n");
		ret = envia (clients[cid].socket, buf_mem, escrits, sbuf, MAX_SBUF);
	}
	/* Alliberem la memria */
	if (munmap (buf_mem, escrits) != 0) {
		perror ("HTTP->XIC_GZIP");
		fprintf (f_err, FORMAT_ERROR, getData (),
				XNODEFINIT, "HTTP->XIC_GZIP", "-");
		mata_thread (pthread_self (), cid);
	}
	/* restaurem la mida si ha fallat alguna cosa */
	if (ret != 0) (*sbuf).mida = mida_ant;
	return ret;
}

/**
 * Funci que intenta enviar el recurs demanat.
 *
 * Parmetres:	cid   - L'identificador d'aquest client
 * 		*sbuf - Punter al buffer de sortida
 *
 * Retorn:	CE403	- Forbidden, si no tenim perms per accedir al fitxer
 * 		CE404	- Not Found, si no trobem el fitxer
 * 		0	- Tot correcte
 * 		-1	- Altres
 */
int cerca_recurs (int cid, buf_out *sbuf)
{
	FILE *fd = NULL;
	struct stat st;
	char fitxer[1500];
	char filebuf[4096];
	int n, ret = 0;
	char *ext;
	c_response resposta;
	char timebuf[100];

	/* Obtenim el path absolut del recurs demanat */
	sprintf (fitxer, "%s%s", config.dir_html, clients[cid].dades->URI);

#ifdef DEBUG
	fprintf (f_err, "Recurs demanat: %s\n", fitxer);
#endif

	if (((char *) strstr(fitxer, "..")) != NULL) return -1; // ;)
	/* Comprovem que existeixi i n'obtenim la mida i la data de l'ltima modificaci entre altres */
	if (stat (fitxer, &st) != 0) {
		if (errno == EACCES)
			return CE403;	// Forbidden - permission denied
		else return CE404;	// Not Fouund
	}
	/* No ens interessen els directoris de moment */
	if ((st.st_mode & S_IFMT) == S_IFDIR) return -1;

	/* Construim les capaleres i la resposta */
	strftime (timebuf, sizeof (timebuf), RFC1123, gmtime (&st.st_mtime));
	sprintf (resposta.LastModified, "Last-Modified: %s\r\n", timebuf);
	sprintf (resposta.Server, "Server: %s\r\n", NOM_SERVIDOR);
	sprintf (resposta.Date, "Date: %s\r\n", getData ());

	/* El tipus MIME */
	ext = (char *) strrchr (&fitxer[0], '.');
	if ((ext == NULL) || (strlen(ext) == 1)) // Si no t extensi o acaba amb '.'
		set_mime (&resposta, NULL);
	else set_mime (&resposta, ++ext);

	sprintf (resposta.estat, "%s %s", clients[cid].dades->versio, HTTP200);
	
	/* Si s una petici CGI comprovem que el recurs sigui executable
	 * i l'executem. Tamb mirem que els path tan per CGI com per 
	 * peticions normals siguin correctes */
	if (CGI) {
		if ((((char *) strstr (fitxer, config.dir_cgi)) != NULL) &&
				(st.st_mode & S_IXGRP))
			return xic_cgi (cid, sbuf, (char *) fitxer);
		else return CE403;
	} else if (((char *) strstr (fitxer, config.dir_cgi)) != NULL)
		return CE403;
	
	fprintf (f_log, LOG_FORMAT, inet_ntoa (clients[cid].adreca_client.sin_addr),
			getData (), log_temp, resposta.estat);

	strcat (resposta.estat, "\r\n");
	
	/* Si s'accepta gzip comprimim tots els tipus de fitxers,
	 * excepte els ja comprimits i les imatges */
	if ((strcmp (clients[cid].dades->AcceptEncoding, "gzip") == 0) &&
			(((char *) strstr (resposta.ContentType, "image")) == NULL) &&
			 (((char *) strstr (resposta.ContentType, "zip")) == NULL) &&
			 (((char *) strstr (resposta.ContentType, "gz")) == NULL)) {
		/* No retornem xic_gzip, ja que si falla encara podem intentar enviar
		 * el fitxer sense comprimir */
		if (xic_gzip (cid, &fitxer[0], st.st_size, &resposta, sbuf) == 0)
			return 0;
	}
	/* Si som aqu no s'accepta compressi (o ha fallat la compressio) */
	/* Construim les capaleres que falten per enviar el fitxer en "clar" */
	sprintf (resposta.ContentLength, "Content-Length: %i\r\n", (int) st.st_size);
	sprintf (resposta.ContentEncoding, "Content-Encoding: identity\r\n");

	/* Preparem la resposta */
	xic_strcat (sbuf, resposta.estat);
	xic_strcat (sbuf, resposta.Date);
	xic_strcat (sbuf, resposta.Server);
	xic_strcat (sbuf, resposta.LastModified);
	xic_strcat (sbuf, resposta.ContentType);
	xic_strcat (sbuf, resposta.ContentLength);
	xic_strcat (sbuf, resposta.ContentEncoding);
	if ((!HTTP11) || (strcasecmp (clients[cid].dades->Connection, "close") == 0))
		xic_strcat (sbuf, "Connection: close\r\n");

	/* Si s un HEAD noms enviem les capaleres
	 * Si s un GET li enviem les capaleres seguit d'un CRLF i del fitxer */
	if (strcmp (clients[cid].dades->metode, "HEAD") == 0) {
		return envia (clients[cid].socket, "\r\n", strlen ("\r\n"), sbuf, MAX_SBUF);
	} else if (strcmp (clients[cid].dades->metode, "GET") == 0) {
		xic_strcat (sbuf, "\r\n");
		if ((fd = fopen (fitxer, "r")) == NULL) {
			fprintf (f_err, FORMAT_ERROR, getData (),
					XRECURS, "HTTP->CERCA_RECURS", "-");
			return -1;
		}
		while ((n = fread (filebuf, 1, 4096, fd)) > 0) {
			ret = envia (clients[cid].socket, filebuf, n, sbuf, MAX_SBUF);
		}
		fclose (fd);
	}
	return ret;
}

/**
 * Funci que comprova si la connexi a la que sock en fa referncia utilitza pipelining.
 *
 * Per fer-ho utilitzem un select amb el timeout a 0 i si es detecta activiat comprovem
 * que realment l'altre extrem ens intenta enviar alguna cosa per mitj d'una consulta
 * recv amb el quart parmetre a MSG_PEEK.
 *
 * Parmetres:	sock - El fd del socket que volem consultar
 *
 * Retorn:	 0 - Si no hi ha pipelining, o no hi ha ms peticions
 * 		 1 - Hi ha pipelining
 * 		-1 - Algun error
 *
 * Nota: Per obtenir una resposta fiable, cal cridar a aquesta funci en el moment oport,
 *       s a dir, quan tenim una resposta a punt per encara no l'hem enviat. Aix
 *       si el client utilitza pipelinig i t ms peticions per fer tindrem dades
 *       pendents de llegir al buffer de recepci del socket.
 */
int comprova_pipelining (int sock)
{
	fd_set fd;
	int n;
	struct timeval t;
	char c = '\r';

	FD_ZERO (&fd);
	FD_SET (sock, &fd);
	for (;;) {
		t.tv_sec = 0;
		t.tv_usec = 0;
		if ((n = select (sock + 1, &fd, NULL, NULL, &t)) == 0) {
			return 0;					// No hi ha pipelining
		}
		if (n == -1) return -1;				// Error
		/* Provem que efectivament hi ha dades per llegir */
		if ((n = recv (sock, &c, 1, MSG_PEEK)) <= 0) return -1;
		if ((c != '\r') && (c != '\n')) return 1;	// Pipeninig
		else recv (sock, &c, 1, MSG_NOSIGNAL);		// Salts adicionals 
	}
}

/**
 * Funci que controla la funcionalitat del protocol HTTP
 *
 * Parmetres:	cid - L'identificador d'aquest client
*/
void http_control (int cid)
{
	fd_set fd;
	buf_out sbuf;
	int n, ce;
	struct timeval tv;
	char c;
	
	/* Reservem espai per emmagatzemar les dades del request del client */
	if ((clients[cid].dades = (c_request *) malloc (sizeof (c_request))) == NULL) {
		fprintf (f_err, FORMAT_ERROR, getData (),
				XNOMEM2, "HTTP->HTTP_CONTROL->CLIENTS.DADES", "-");
		return;
	}

	sbuf.mida = 0;
	
	FD_ZERO (&fd);
	FD_SET (clients[cid].socket, &fd);

	for (;;) {
		/* Timeout de la connexi */
		// Cal resetejar el temps cada vegada ja que el select el modifica
		tv.tv_sec = config.timeout;
		tv.tv_usec = 0;
		reset_request(cid);
		if ((n = select (clients[cid].socket + 1, &fd, NULL, NULL, &tv)) <= 0) {
			// timeout (o error)
			FD_CLR (clients[cid].socket, &fd);	// s necessari ??
			return;
		}
		/* Mirem que l'"activitat" no sigui TANCAR */
		if ((n = recv (clients[cid].socket, &c, 1, MSG_PEEK)) <= 0) {
			FD_CLR (clients[cid].socket, &fd);	// ??
			return;
		}
		/* Procedim a tractar el request */
		if ((ce = deco_request (cid)) == 0)
			ce = cerca_recurs (cid, &sbuf);
		switch (ce) {
			case -1: 	// Acabem
				return;
				break;
			case 0:		// Ja est tot fet
				break;
			default:
				prepara_error (ce, cid, &sbuf);
		}
		/* No es soporta la persistncia en connexions HTTP/1.0 */
		if ((!HTTP11) || (strcasecmp (clients[cid].dades->Connection, "close") == 0)) {
			FD_CLR (clients[cid].socket, &fd);	// ??
			return;
		}
	}
}

