/**
 * Definici d'estructures per al protocol HTTP
 */

#define MAX_SBUF 40960

/* Aquesta mida ha estat calculada sumant la mida mxima
 * que s'encuar al buffer de sortida "sense control", i
 * bsicament sn la suma de l'estat i les capaleres
 */
#define MIDA_MAX_RESPONSE 800

/* Codis d'estat */
#define CE200 0
#define CE400 1
#define CE501 2
#define CE414 3
#define CE505 4
#define CE404 5
#define CE403 6
#define CE500 7
#define CE411 8

/* Format de la data */
#define RFC1123 "%a, %d %b %Y %H:%M:%S GMT"

/* Missages d'estat del servidor 
 * (Segurament la majoria no s'utilitzaran en aquest servidor)*/
#define HTTP100	"100 Continue"
#define HTTP101 "101 Switching Protocols"

#define HTTP200 "200 OK"
#define HTTP201 "201 Created"
#define HTTP202 "202 Accepted"
#define HTTP203 "203 Non-Authoritative Information"
#define HTTP204 "204 No Content"
#define HTTP205 "205 Reset Content"
#define HTTP206 "206 Partial Content"

#define HTTP300 "300 Multiple Choices"
#define HTTP301 "301 Moved Permanently"
#define HTTP302 "302 Found"
#define HTTP303 "303 See Other"
#define HTTP304 "304 Not Modified"
#define HTTP305 "305 Use Proxy"
// 306 reservat
#define HTTP307 "307 Temporary Redirect"

#define HTTP400 "400 Bad Request"
#define HTTP401 "401 Unauthorized"
#define HTTP402 "402 Payment Required"
#define HTTP403 "403 Forbidden"
#define HTTP404 "404 Not Found"
#define HTTP405 "405 Method Not Allowed"
#define HTTP406 "406 Not Acceptable"
#define HTTP407 "407 Proxy Authentication Required"
#define HTTP408 "408 Request Timeout"
#define HTTP409 "409 Conflict"
#define HTTP410 "410 Gone"
#define HTTP411 "411 Length Required"
#define HTTP412 "412 Precondition Failed"
#define HTTP413 "413 Request Entity Too Large"
#define HTTP414 "414 Request-URI Too Long"
#define HTTP415 "415 Unsuported Media Type"
#define HTTP416 "416 Request Range Not Satisfiable"
#define HTTP417 "417 Expectation Failed"

#define HTTP500 "500 Internal Server Error"
#define HTTP501 "501 Not Implemented"
#define HTTP502 "502 Bad Gateway"
#define HTTP503 "503 Service Unavailable"
#define HTTP504 "504 Gateway Timeout"
#define HTTP505 "505 HTTP Version Not Suported"

/* Estructura dels request */
typedef struct {
	char metode[16];
	char versio[16];
	char URI[1024];
	char Host[64];
	char AcceptEncoding[16];
	char get_p[1024];
	int ContentLength;
	char Connection[64];
} c_request;

/* Estructura dels response */
typedef struct {
	char estat[64];
	char Date[128];
	char Server[128];
	char LastModified[128];
	char ContentLength[128];
	char ContentType[128];
	char ContentEncoding[64];
} c_response;

/* Buffer de sortida, s necessari que sigui aix ja que en molts
 * casos si afegeixen dades binries, el que provoca que els bytes
 * amb valor 0 ens juguin males passades a l'hora de saber-ne la mida
 */
typedef struct {
	char sbuf[MAX_SBUF];
	int mida;
} buf_out;

/* Declaraci d'algunes funcions */
char *getData (void);
int llegir_linia (int sock, char *buf, int mida);
int comprova_pipelining (int sock);
void http_control (int cid);
