/**
 * Fitxer amb les tasques del servidor
 *
 * Autor: Miquel Fontanals Rofes
 *
 */

#include "main.h"

/* Variables per emmagatzemar l'EUID i el UID respectivament (veure init ())*/
uid_t e_uid_inicial;
uid_t r_uid;

void prepara_senyals ();
void allibera_connexio (int cid);

/**
 * Funci que inicialitza el fitxer d'errors
 */
void err_init ()
{
#ifndef DEBUG
	FILE *aux;

	seteuid (e_uid_inicial);
	if ((aux = freopen (config.f_err, "w", stderr)) == NULL) {
		char msg[512];
		sprintf (msg,
		"No s'ha pogut obrir el fitxer d'errors %s !!!\n
		S'utilitzar la sortida estndards d'errors stderr.",
		config.f_err);
				
		fprintf (f_err, FORMAT_ERROR, getData (),
				XOLOG, "LOGS->ERR_INIT", msg);
	} else f_err = aux;
	seteuid (r_uid);
#endif
}

/**
 * Funcio que inicialitza el fitxer per logar els accessos
 */
void log_init ()
{
	char timebuf[100];
	struct timeval ara;
	struct timezone tzona;
#ifndef DEBUG2
	FILE *aux;
	seteuid (e_uid_inicial);
	if ((aux = fopen (config.f_log, "a")) == NULL) {
		char msg[512];
		if ((aux = fopen ("/dev/null", "w")) == NULL) {
			f_log = stdout;
			sprintf (msg, "ATENCI !!!\n
			S'utilitzar la sortida estndard stdout per logar els accessos !");
			fprintf (f_err, FORMAT_ERROR, getData (),
					XOLOG, "LOGS->LOG_INIT", msg);
			return;
		}
		sprintf (msg, "ATENCI !!! No es pot obrir el fitxer de log %s\n
				NO es logaran els accessos !", config.f_log);
		fprintf (f_err, FORMAT_ERROR, getData (),
				XOLOG, "LOGS->INITOG", msg);
	} else f_log = aux;
	seteuid (r_uid);
#endif
	gettimeofday (&ara, &tzona);
	strftime (timebuf, sizeof (timebuf), "%c", localtime (&ara.tv_sec));
	fprintf (f_log, "\n\n\tLog d'accessos inicialitzat a %s\n\n", timebuf);
	fflush (f_log);
}

/**
 * Funci que executaran els threads dinmics
 *
 * Parmetres:
 * 		- cid: punter a l'enter que representa l'id del thread
 */
void *fil (int *cid)
{
	clients[*cid].fil = pthread_self();
	http_control (*cid);
	allibera_connexio (*cid);
	pthread_exit (NULL);
}

/**
 * Funci que executaran els threads esttics
 *
 * Parmetres:
 * 		- cid: punter a l'enter que representa l'id del thread
 */
void *fil_estatic (int *cid)
{
	clients[*cid].fil = pthread_self();
	for (;;) {
		pthread_mutex_lock (&mutex[*cid]);
		pthread_cond_wait (&cond[*cid], &mutex[*cid]);
		pthread_mutex_unlock (&mutex[*cid]);
		http_control (*cid);
		allibera_connexio (*cid);
	}
}

/**
 * Funci que crea els threads esttics especificats al fitxer de
 * configuraci.
 */
void crea_fils_estatics ()
{
	int i;
	for (i = 0; i < config.min_threads; i++) {
		clients[i].id = i;
		pthread_create (&fils[i], NULL, (void *) &fil_estatic, (void *) &clients[i].id);
		pthread_detach (fils[i]);
	}
}

/**
 * Funci que retorna un identificador lliure per a les connexions
 * entrants dels clients.
 *
 * Retorn:
 * 		- enter: l'id per la nova connexi.
 */
int getId () {
	int id;
	for (id = 0; id < config.max_threads; id++) {
		if (clients[id].socket == 0) return id;
	}
	return -1;
}

/**
 * Funci que resta a l'espera de rebre connexions pel socket
 * passat per parmetre.
 *
 * Parmetres:
 * 		- sock: el socket del servidor que acceptar les connexions
 */
void escolta (int sock)
{
	int r, sc, id = 0;
	struct sockaddr_in adreca_client;
	size_t mida_client;
	/* Preparem el socket per acceptar connexions */
	if ((r = listen (sock, 50)) == -1) {
		perror ("SERVER->ESCOLTA->LISTEN");
		fprintf (f_err, FORMAT_ERROR, getData (), XLISTEN, "SERVER->ESCOLTA->LISTEN", "-");
		exit (-1);
	}
	/* Bucle infinit per acceptar connexions */
	for (;;) {
		mida_client = sizeof (struct sockaddr_in);
		/* Acceptem les connexions */
		if ((sc = accept (sock, (struct sockaddr *) &adreca_client, &mida_client)) == -1) {
			perror ("SERVER->ESCOLTA->ACCEPT");
			fprintf ( f_err, FORMAT_ERROR, getData (), XACCEPT,
					"SERVER->ESCOLTA->ACCEPT", "-");
			exit (-1);
		}
		/* Obtenim un id */
		if ((id = getId()) == -1) {
			fprintf (f_err, FORMAT_ERROR, getData (), XMASSACON, "SERVER->ESCOLTA", "-");
			close (sc);
			continue;
		}
		/* Inicialitzem els valors bsics */
		clients[id].id = id;
		clients[id].socket = sc;
		memmove (&clients[id].adreca_client, &adreca_client, mida_client);
		/* Tenim threads esttics lliures ? */
		if (id < config.min_threads) {
			pthread_mutex_lock (&mutex[id]);
			pthread_cond_signal (&cond[id]);
			pthread_mutex_unlock (&mutex[id]);
		} else {	/* En creem un dinmicament */
			pthread_create (&fils[id], NULL, (void *) &fil, (void *) &clients[id].id);
			pthread_detach (fils[id]);
		}
	}
}

/**
 * Funci que crea un socket i li assigna l'adrea del host local
 * 
 * Retorn:
 * 		- enter: El fd del socket creat
 */
int crea_socket ()
{
	struct hostent *host;
	int i = 0;
	/* Obrim el socket */
	if ((server_sock = socket (AF_INET, SOCK_STREAM, 0)) == -1) {
		perror ("SERVER->CREA_SOCKET->SOCKET");
		fprintf (f_err, FORMAT_ERROR, getData (), XSOCKET, "SERVER->CREA_SOCKET->SOCKET", "-");
		exit (-1);
	}
	/* Obtenim la direcci IP a partir d'una cerca DNS seguint l'ordre establert a /etc/hosts.conf */
	config.addr.sin_family = AF_INET;
	if ((host = (struct hostent *) gethostbyname (config.nom_host)) == NULL) {
		perror ("SERVER->CREA_SOCKET->GETHOSTGYNAME");
		fprintf (f_err, FORMAT_ERROR, getData (), XDNS, "SERVER->CREA_SOCKET->GETHOSTBYNAME",
				config.nom_host);
		exit (-1);
	}
	memmove ((unsigned char *) &config.addr.sin_addr, host->h_addr, host->h_length);
	config.addr.sin_port = htons (config.port);
	/* Publiquem el socket (5 intents) */
	seteuid (e_uid_inicial);	// Pels ports reservats
	while (bind (server_sock, (struct sockaddr *) &config.addr, sizeof (struct sockaddr)) == -1) {
		i++;
		perror ("SERVER->CREA_SOCKET->BIND");
		fprintf (f_err, FORMAT_ERROR, getData (), XBIND, "SERVER->CREA_SOCKET->BIND", "-");
		if (i >= 5) exit (-1);
		sleep (5);
	}
	seteuid (r_uid);
	return 0;
}

/**
 * Funci que llegeix el fitxer de configuraci i omple els valors
 * de les variables de l'estructura config.
 */
void llegir_config ()
{
	FILE *fc = NULL;
	char *var, *val, *linia = NULL;
	size_t numchars = 0;

	fprintf (f_log, "Llegint configuraci: ");
	/* Obrim el fitxer */
	if ((fc = fopen (config.fitxer_config, "r")) == NULL)
		fprintf (f_err, FORMAT_ERROR, getData (), XCONFIG, "SERVER->LLEGIR_CONFIG->FOPEN1",
				config.fitxer_config);
	/* Ho intentem amb el nostre etc */
	if (fc == NULL) {
		sprintf (config.fitxer_config, "../etc/xichttpd.conf");
		if ((fc = fopen (config.fitxer_config, "r")) == NULL)
			fprintf (f_err, FORMAT_ERROR, getData (), XCONFIG,
					"SERVER->LLEGIR_CONFIG->FOPEN2", config.fitxer_config);
	}
	/* Si tampoc no el trobem intenem cercar-lo al directori actual */
	if (fc == NULL) {
		sprintf (config.fitxer_config, "xichttpd.conf");
		if ((fc = fopen (config.fitxer_config, "r")) == NULL) {
			perror ("SERVER->LLEGIR_CONFIG->FOPEN2");
			fprintf (f_err, FORMAT_ERROR, getData (), XCONFIG2,
					"SERVER->LLEGIR_CONFIG->FOPEN3", config.fitxer_config);
			exit (-1);
		}
	}
	fprintf (f_log, "Utilitzant %s\n", config.fitxer_config);
	/* El fitxer existeix */
	while (getline (&linia, &numchars, fc) != -1) {
		if (linia[0] != '#') {
			if ((var = (char *) strtok (linia, "=")) == NULL) {
				fprintf (f_err, FORMAT_ERROR, getData (), 
						XCONFIG3, "SERVER->LLEGIR_CONFIG->LINIA", linia);
				exit (-1);
			}
			var[strlen (var) - 1] = '\0';
			if ((val = (char *) strtok (NULL, "\" ")) != NULL) {
				if ((strcmp (var, "DIR_BASE") == 0) && (strlen (val) < 255))
					sprintf (config.dir_base, "%s", val);
				else if ((strcmp (var, "DIR_CGI") == 0) && (strlen (val) < 255))
					sprintf (config.dir_cgi, "%s", val);
				else if ((strcmp (var, "DIR_HTML") == 0) && (strlen (val) < 255))
					sprintf (config.dir_html, "%s", val);
				else if ((strcmp (var, "FITXER_ERR") == 0) && (strlen (val) < 255))
					sprintf (config.f_err, "%s", val);
				else if ((strcmp (var, "FITXER_LOG") == 0) && (strlen (val) < 255))
					sprintf (config.f_log, "%s", val);
				else if ((strcmp (var, "MAX_CLIENTS") == 0) && (strlen (val) <= 5))
					config.max_threads = atoi (val);
				else if ((strcmp (var, "MIME_TYPES") == 0) && (strlen (val) < 255))
					sprintf (config.mime_types, "%s", val);
				else if ((strcmp (var, "MIN_THREADS") == 0) && (strlen (val) <= 5))
					config.min_threads = atoi (val);
				else if ((strcmp (var, "NOM_HOST") == 0) && (strlen (val) < 64))
					sprintf (config.nom_host, "%s", val);
				else if ((strcmp (var, "PORT") == 0) && (strlen (val) <= 5))
					config.port = atoi (val);
				else if ((strcmp (var, "SERVER_TIMEOUT") == 0) && (strlen (val) <= 5))
					config.timeout = (atoi (val) > 0 ? atoi (val): 0) ;
			}
		}
	}
	fclose (fc);
	/* Comprovem que tot sigui correcte */
	if ((config.min_threads == 0) ||
			(config.max_threads == 0) ||
			(config.min_threads > config.max_threads)) {
		fprintf (f_err, FORMAT_ERROR, getData (),
				XCONFIG3, "SERVER->LLEGIR_CONFIG->COMPROVACIONS",
	"Ni MIN_THREADS ni MAX_CLIENTS no poden ser 0, i a ms MIN_THREADS ha de ser menor o igual que MAX_CLIENTS");
		exit (-1);
	}
	if ((strlen (config.dir_base) == 0)
			|| (strlen (config.dir_cgi) == 0)
			|| (strlen (config.dir_html) == 0)
			|| (strlen (config.f_err) == 0)
			|| (strlen (config.f_log) == 0)
			|| (strlen (config.mime_types) == 0)
			|| (config.timeout == 0)) {
		fprintf (f_err, FORMAT_ERROR, getData (),
				XCONFIG3, "SERVER->LLEGIR_CONFIG->COMPROVACIOS",
				"Comprovi el fitxer de configuraci i respecti els espais del mateix");
		exit (-1);
	}
}

/**
 * Funci que carrega els tipus mime definits al fitxer etc/mime.types
 */
void carrega_mime ()
{
	FILE *mt = NULL;
	char *linia = NULL;
	size_t numchars = 0;
	mime *actual;
	int i = 0;
	char *tm, *ext;

	/* Inicialitzem l'estructura per als tipus mime */
	if ((mime_types = (mime *) malloc (sizeof (mime))) == NULL) {
		fprintf (f_err, FORMAT_ERROR, getData (), XNOMEM, "SERVER->CARREGA_MIME->MALLOC", "-");
		exit (-1);
	}
	mime_types->next = NULL;
	actual = mime_types;
	/* Intentem obrir el fitxer que cont els tipus mime */
	if ((mt = fopen (config.mime_types, "r")) == NULL) {
		perror ("SERVER->CARREGA_MIME->FOPEN");
		fprintf (f_err, FORMAT_ERROR, getData (), XMIME, "SERVER->CARREGA_MIME->FOPEN",
				config.mime_types);
		exit (-1);
	}
	/* El parsejem i omplim la llista mime_types */
	while (getline (&linia, &numchars, mt) != -1) {
		if (linia[0] != '#') {
			if (((tm = (char *) strtok (linia, "\t \r\n")) != NULL) &&
					(strlen (tm) < 64)) {
				while (((ext = (char *) strtok (NULL, "\t \r\n")) != NULL) &&
						(strlen (ext) < 16)) {
					sprintf (actual->tipus_mime, "%s", tm);
					sprintf (actual->ext, "%s", ext);
					i++;
					if ((actual->next = (mime *) malloc (sizeof (mime))) == NULL) {
						fprintf (f_err, FORMAT_ERROR, getData (), XNOMEM,
							"SERVER->CARREGA_MIME->MALLOC", "-");
						exit (-1);	
					}
					actual = actual->next;
				}
			}
		}
	}
	actual = NULL; // fi de llista
	if (i == 0) {
		fprintf (f_err, FORMAT_ERROR, getData (), XMIME, "SERVER->CARREGA_MIME->COMPROVA",
				"No s'han pogut carregar els tipus mime");
		exit (-1);
	}
#ifdef DEBUG
	for (actual = mime_types; actual->next != NULL; actual = actual->next) {
		fprintf (f_err, "%s - %s\n", actual->ext, actual->tipus_mime);
	}
#endif
	fprintf (f_err, "%i tipus mime carregats\n", i);
}
/**
 * Funci que reseteja els parmetres de les connexions.
 * s cridada cada vegada que s'acava la comunicaci amb un client.
 *
 * Parmetres:
 * 		- cid: L'identificador de la connexi acabada
 */
void allibera_connexio (int cid)
{
	memset (&clients[cid].adreca_client, 0, sizeof (struct sockaddr_in));
	clients[cid].ctemps = 0;
	clients[cid].atemps = 0;
	free (clients[cid].dades);
	close (clients[cid].socket);
	fflush (NULL);
	clients[cid].socket = 0;
}

/**
 * Funci que finalitza un thread o el propi servidor
 * Normalment ser cridada per la rutina d'atenci dels signals (tracta_senyal)
 *
 * Parmetres:
 * 		- fil: Identificador del thread que volem finalitzar
 * 		- cid: L'id de la connexi que volem finalitzar,
 * 			si el seu valor s -1 s perqu venim d'un signal
 */
void mata_thread (pthread_t fil, int cid)
{
	int i;
	/* Finalitzem el servidor */
	if (pthread_equal (fil, pare)) {
		close (server_sock);
		fflush (NULL);
		fclose (f_err);
		fclose (f_log);
		exit (0);
	}
	/* Quasi mai passar ja que podem cridar directament la funci
	 * allibera_connexio quan convingui.*/
	if (cid > -1) {
		allibera_connexio (cid);
		return;
	}
	/* venim d'un signal i no sabem quin thread alliberar
	 * Si s un dels esttics sortim */
	for (i = 0; i < config.min_threads; i++) {
		if (pthread_equal (fil, clients[i].fil)) {
			mata_thread (pare, -1);
		}
	}
	/* Sino l'alliberem i el tanquem */
	for (i = config.min_threads; i < config.max_threads; i++) {
		if (pthread_equal (fil, clients[i].fil)) {
			allibera_connexio (i);
			return;
		}
	}
	/* Si malgrat tot hem arribat aqu sortim */
	fprintf (f_err, FORMAT_ERROR, getData (), XNODEFINIT, "SERVER->MATA_THREAD", "????");
	exit (-1);
}

/**
 * Rutina d'atenci als signals.
 *
 * Parmetre:
 * 		- snyl: El signal rebut
 */
void tracta_senyal (int snyl)
{
	switch (snyl) {
		case SIGHUP:
			fprintf (f_err, FORMAT_ERROR, getData (),
					XSIGHUP, "SERVER->TRACTA_SENYAL", "-");
			break;
		case SIGINT:
			fprintf (f_err, FORMAT_ERROR, getData (), 
					XSIGINT, "SERVER->TRACTA_SENYAL", "-");
			break;
		case SIGSEGV:
			fprintf (f_err, FORMAT_ERROR, getData (),
					XSIGSEGV, "SERVER->TRACTA_SENYAL", "-");
			break;
		case SIGPIPE:
			fprintf (f_err, FORMAT_ERROR, getData (),
					XSIGPIPE, "SERVER->TRACTA_SENYAL", "-");
			break;
		case SIGTERM:
			fprintf (f_err, FORMAT_ERROR, getData (),
					XSIGTERM, "SERVER->TRACTA_SENYAL", "-");
			break;
		default:
			fprintf (f_err, FORMAT_ERROR, getData (),
					XSIG, "SERVER->TRACTA_SENYAL", "-");
	}
	// De moment tanquem el servidor
	mata_thread (pare, -1);
}

/**
 * Funci que assigna la rutina d'atenci als signals que ens interessen
 */
void prepara_senyals ()
{
	signal (SIGHUP, (void *) tracta_senyal);
	signal (SIGINT, (void *) tracta_senyal);
	signal (SIGSEGV, (void *) tracta_senyal);
	signal (SIGPIPE, (void *) tracta_senyal);
	signal (SIGTERM, (void *) tracta_senyal);
}

/**
 * Funci que inicialitza el servidor.
 */
void init ()
{
	int i;

	/* Guardem l'usuari efectiu i el real */
	e_uid_inicial = geteuid ();
	r_uid = getuid ();
	
	/* Canviem a l'usuari real, les funcions que necessitin
	 * els privilegis ja fan el canvi internament */
	seteuid (r_uid);

	/* Sortides per defecte */
	f_err = stderr;
	f_log = stdout;
	
	fprintf (f_log, "Arrencant Servidor: %s...\n", NOM_SERVIDOR);
	prepara_senyals ();
	pare = pthread_self();
	llegir_config ();
	carrega_mime ();
	err_init ();
	log_init ();
	
	/* Guardem espai per la taula de clients */
	if ((clients = (connexio *) calloc (config.max_threads, sizeof (connexio))) == NULL) {
		perror ("SERVER->INIT->CLIENTS");
		fprintf (f_err, FORMAT_ERROR, getData (), XNOMEM, "SERVER->INIT->CLIENTS", "-");
		exit (-1);
	}
	
	/* Tamb per la taula de fils */
	if ((fils = (pthread_t *) calloc (config.max_threads, sizeof (pthread_t))) == NULL) {
		perror ("SERVER->INIT->FILS");
		fprintf (f_err, FORMAT_ERROR, getData (), XNOMEM, "SERVER->INIT->FILS", "-");
		exit (-1);
	}
	
	/* Taula de condicionals */
	if ((cond = (pthread_cond_t *) calloc (config.min_threads, sizeof (pthread_cond_t))) == NULL) {
		perror ("SERVER->INIT->COND");
		fprintf (f_err, FORMAT_ERROR, getData (), XNOMEM, "SERVER->INIT->COND", "-");
		exit (-1);
	}
	
	/* Taula de mutex */
	if ((mutex = (pthread_mutex_t *) calloc (config.min_threads, sizeof (pthread_mutex_t))) == NULL) {
		perror ("SERVER->INIT->MUTEX");
		fprintf (f_err, FORMAT_ERROR, getData (), XNOMEM, "SERVER->INIT->MUTEX", "-");
		exit (-1);
	}
	
	/* Inicialitzem les variables condicionals i mutex */
	for (i = 0; i < config.min_threads; i++) {
		pthread_mutex_init (&mutex[i], NULL);
		pthread_cond_init (&cond[i], NULL);
	}
	
	crea_socket ();

	/* Ens convertim en daemon */
	switch (fork()) {
		case -1: exit (-1);
		case 0: break;
		default: exit (0);
	}
	setsid ();	/* Nova sessi */
	
	crea_fils_estatics ();			
	escolta (server_sock);
	close (server_sock); // No hi arribarem mai
}
